#include "Renderer.hpp"

#include <algorithm>
#include <ranges>
#include <vector>

const std::vector<const char*> g_layers{
#ifdef RENDERER_VULKAN_DEBUG
    "VK_LAYER_KHRONOS_validation"
#endif
};

const std::vector<const char*> g_device_extensions{ VK_KHR_SWAPCHAIN_EXTENSION_NAME };

[[nodiscard]]
static auto create_instance() -> vk::UniqueInstance
{
    constexpr static vk::ApplicationInfo application_info{
        .apiVersion = VK_API_VERSION_1_0,   // version 1.0 is always supported
    };

    // Ideally we should check for the availability of layers and extensions.
    // However, vk::createInstanceUnique is going to throw when something is unavailable
    // (given that vulkan exceptions are enabled)

    const vk::InstanceCreateInfo create_info{
        .pApplicationInfo    = &application_info,
        .enabledLayerCount   = static_cast<uint32_t>(g_layers.size()),
        .ppEnabledLayerNames = g_layers.data(),
        .enabledExtensionCount =
            static_cast<uint32_t>(Window::vulkan_instance_extensions().size()),
        .ppEnabledExtensionNames = Window::vulkan_instance_extensions().data(),
    };

    return vk::createInstanceUnique(create_info);
}

[[nodiscard]]
static auto supports_required_extensions(const vk::PhysicalDevice physical_device) -> bool
{
    const auto properties{ physical_device.enumerateDeviceExtensionProperties() };

    return std::ranges::all_of(g_device_extensions, [&](const char* extension_name) {
        return std::ranges::find_if(
                   properties,
                   [extension_name](const char* property_extension_name) {
                       return std::strcmp(extension_name, property_extension_name) == 0;
                   },
                   [](const vk::ExtensionProperties& extension_properties
                   ) -> const char* { return extension_properties.extensionName; }
               )
            != std::end(properties);
    });
}

[[nodiscard]]
static auto find_queue_family_index(
    const vk::PhysicalDevice physical_device,
    const vk::SurfaceKHR     surface
) -> std::optional<uint32_t>
{
    uint32_t index{};
    for (const auto& properties : physical_device.getQueueFamilyProperties()) {
        if (properties.queueFlags & vk::QueueFlagBits::eGraphics
            && physical_device.getSurfaceSupportKHR(index, surface))
        {
            return index;
        }
        index++;
    }
    return std::nullopt;
}

[[nodiscard]]
static auto
    choose_physical_device(const vk::Instance instance, const vk::SurfaceKHR surface)
        -> vk::PhysicalDevice
{
    std::vector<vk::PhysicalDevice> physical_devices{};
    std::ranges::remove_copy_if(
        instance.enumeratePhysicalDevices(),
        std::back_inserter(physical_devices),
        [surface](vk::PhysicalDevice physical_device) {
            return !(
                supports_required_extensions(physical_device)
                && find_queue_family_index(physical_device, surface).has_value()
            );
        }
    );

    if (std::ranges::empty(physical_devices)) {
        throw std::runtime_error{ "No Vulkan physical device is available." };
    }

    const auto ranked_devices_view{
        physical_devices | std::views::transform([](vk::PhysicalDevice physical_device) {
            switch (physical_device.getProperties().deviceType) {
                case vk::PhysicalDeviceType::eDiscreteGpu:
                    return std::make_pair(physical_device, 2);
                case vk::PhysicalDeviceType::eIntegratedGpu:
                    return std::make_pair(physical_device, 1);
                default: return std::make_pair(physical_device, 0);
            }
        })
    };
    std::vector ranked_devices(ranked_devices_view.begin(), ranked_devices_view.end());

    std::ranges::sort(
        ranked_devices, std::ranges::greater{}, &std::pair<vk::PhysicalDevice, int>::second
    );

    return ranked_devices.front().first;
}

[[nodiscard]]
static auto
    create_device(const vk::PhysicalDevice physical_device, const vk::SurfaceKHR surface)
        -> vk::UniqueDevice
{
    constexpr static std::array queue_priorities{ 1.f };

    const vk::DeviceQueueCreateInfo queue_create_info{
        .queueFamilyIndex = find_queue_family_index(physical_device, surface).value(),
        .queueCount       = 1,
        .pQueuePriorities = queue_priorities.data(),
    };

    const vk::DeviceCreateInfo device_create_info{
        .queueCreateInfoCount    = 1,
        .pQueueCreateInfos       = &queue_create_info,
        .enabledExtensionCount   = static_cast<uint32_t>(g_device_extensions.size()),
        .ppEnabledExtensionNames = g_device_extensions.data(),
    };

    return physical_device.createDeviceUnique(device_create_info);
}

Renderer::Renderer(const Window& window)
    : m_instance{ create_instance() },
      m_surface{ window.create_vulkan_surface(m_instance.get()) },
      m_physical_device{ choose_physical_device(m_instance.get(), m_surface.get()) },
      m_queue_family_index{
          find_queue_family_index(m_physical_device, m_surface.get()).value()
      },
      m_device{ create_device(m_physical_device, m_surface.get()) },
      m_queue{ m_device->getQueue(m_queue_family_index, 0) },
      m_swapchain{ Swapchain::create(
          m_surface.get(),
          m_physical_device,
          m_queue_family_index,
          m_device.get(),
          window.framebuffer_size()
      ) }
{}
