#include "Swapchain.hpp"

[[nodiscard]]
static auto choose_swapchain_surface_format(
    const vk::SurfaceKHR     surface,
    const vk::PhysicalDevice physical_device
) -> vk::SurfaceFormatKHR
{
    const auto available_surface_formats{ physical_device.getSurfaceFormatsKHR(surface) };

    for (const auto& surface_format : available_surface_formats) {
        if (surface_format.format == vk::Format::eB8G8R8A8Srgb
            && surface_format.colorSpace == vk::ColorSpaceKHR::eSrgbNonlinear)
        {
            return surface_format;
        }
    }

    for (const auto& surface_format : available_surface_formats) {
        if (surface_format.colorSpace == vk::ColorSpaceKHR::eSrgbNonlinear) {
            return surface_format;
        }
    }

    return available_surface_formats.front();
}

[[nodiscard]]
static auto get_min_image_count(const vk::SurfaceCapabilitiesKHR& surface_capabilities
) noexcept -> uint32_t
{
    uint32_t image_count = surface_capabilities.minImageCount + 1;
    if (surface_capabilities.maxImageCount > 0
        && image_count > surface_capabilities.maxImageCount)
    {
        image_count = surface_capabilities.maxImageCount;
    }

    return image_count;
}

[[nodiscard]]
static auto create_swapchain(
    const vk::SurfaceKHR              surface,
    const uint32_t                    queue_family_index,
    const vk::Device                  device,
    const vk::SurfaceCapabilitiesKHR& surface_capabilities,
    const vk::Extent2D                extent,
    const vk::SurfaceFormatKHR        surface_format
) -> vk::UniqueSwapchainKHR
{
    const std::array queue_family_indices{ queue_family_index };

    const vk::SwapchainCreateInfoKHR create_info{
        .surface               = surface,
        .minImageCount         = get_min_image_count(surface_capabilities),
        .imageFormat           = surface_format.format,
        .imageColorSpace       = surface_format.colorSpace,
        .imageExtent           = extent,
        .imageArrayLayers      = 1,
        .imageUsage            = vk::ImageUsageFlagBits::eColorAttachment,
        .imageSharingMode      = vk::SharingMode::eExclusive,
        .queueFamilyIndexCount = static_cast<uint32_t>(queue_family_indices.size()),
        .pQueueFamilyIndices   = queue_family_indices.data(),
        .preTransform          = surface_capabilities.currentTransform,
        .compositeAlpha        = vk::CompositeAlphaFlagBitsKHR::eOpaque,
        .presentMode           = vk::PresentModeKHR::eFifo,
        .clipped               = vk::True
    };

    return device.createSwapchainKHRUnique(create_info);
}

auto Swapchain::choose_extent(
    const vk::Extent2D&               framebuffer_size,
    const vk::SurfaceCapabilitiesKHR& surface_capabilities
) -> vk::Extent2D
{
    if (surface_capabilities.currentExtent.width != std::numeric_limits<uint32_t>::max())
    {
        return surface_capabilities.currentExtent;
    }

    return vk::Extent2D{
        .width = std::clamp(
            static_cast<uint32_t>(framebuffer_size.width),
            surface_capabilities.minImageExtent.width,
            surface_capabilities.maxImageExtent.width
        ),
        .height = std::clamp(
            static_cast<uint32_t>(framebuffer_size.height),
            surface_capabilities.minImageExtent.height,
            surface_capabilities.maxImageExtent.height
        ),
    };
}

auto Swapchain::create(
    const vk::SurfaceKHR     surface,
    const vk::PhysicalDevice physical_device,
    const uint32_t           queue_family_index,
    const vk::Device         device,
    const vk::Extent2D&      framebuffer_size
) -> std::optional<Swapchain>
{
    const auto surface_capabilities{ physical_device.getSurfaceCapabilitiesKHR(surface) };

    const auto extent = choose_extent(framebuffer_size, surface_capabilities);
    if (extent.width == 0 || extent.height == 0) {
        return std::nullopt;
    }

    const auto surface_format{ choose_swapchain_surface_format(surface, physical_device) };

    auto swapchain{ create_swapchain(
        surface, queue_family_index, device, surface_capabilities, extent, surface_format
    ) };

    return Swapchain{ extent, surface_format.format, std::move(swapchain) };
}

auto Swapchain::get() const noexcept -> vk::SwapchainKHR
{
    return m_swapchain.get();
}

auto Swapchain::extent() const noexcept -> vk::Extent2D
{
    return m_extent;
}

auto Swapchain::format() const noexcept -> vk::Format
{
    return m_format;
}

Swapchain::Swapchain(
    const vk::Extent2D&      extent,
    vk::Format               format,
    vk::UniqueSwapchainKHR&& swapchain
) noexcept
    : m_extent{ extent },
      m_format{ format },
      m_swapchain{ std::move(swapchain) }
{}
